%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

-module(odbc).

-behaviour(gen_server).


%% External exports------------------------------------
%% New Basic API
-export([sqlBindColumn/3, sqlBindColumn/4]).
-export([sqlCloseCursor/1, sqlCloseCursor/2]).
-export([sqlConnect/4, sqlConnect/5]).
-export([sqlDescribeCol/2, sqlDescribeCol/3]).
-export([sqlDisConnect/1, sqlDisConnect/2]).
-export([sqlExecDirect/2, sqlExecDirect/3]).
-export([sqlFetch/1, sqlFetch/2]).
-export([sqlNumResultCols/1, sqlNumResultCols/2]).
-export([sqlRowCount/1, sqlRowCount/2]).
-export([columnRef/0]).
-export([readData/2, readData/3]).

%% New Utility API
-export([erl_connect/2, erl_connect/3, erl_connect/4, erl_connect/5]).
-export([erl_executeStmt/2, erl_executeStmt/3]).
-export([erl_disconnect/2, erl_disconnect/3]).

%% Old Basic API
-export([sql_alloc_handle/3, sql_alloc_handle/4]).
-export([sql_bind_col/4, sql_bind_col/5]).
-export([sql_close_cursor/2, sql_close_cursor/3]).
-export([sql_connect/5, sql_connect/6]).
-export([sql_describe_col/4, sql_describe_col/5]).
-export([sql_disconnect/2, sql_disconnect/3]).
-export([sql_driver_connect/5, sql_driver_connect/6]).
-export([sql_end_tran/4, sql_end_tran/5]).
-export([sql_exec_direct/3, sql_exec_direct/4]).
-export([sql_fetch/2, sql_fetch/3]).
-export([sql_free_handle/3, sql_free_handle/4]).
-export([sql_get_connect_attr/4, sql_get_connect_attr/5]).
-export([sql_get_diag_rec/5, sql_get_diag_rec/6]).
-export([sql_num_result_cols/2, sql_num_result_cols/3]).
-export([sql_row_count/2, sql_row_count/3]).
-export([sql_set_connect_attr/5, sql_set_connect_attr/6]).
-export([sql_set_env_attr/5, sql_set_env_attr/6]).
-export([alloc_buffer/3, alloc_buffer/4]).
-export([dealloc_buffer/2, dealloc_buffer/3]).
-export([read_buffer/2, read_buffer/3]).

%% Old Utility API
-export([init_env/1, init_env/2]).
-export([connect/3, connect/4, connect/5, connect/6]).
-export([execute_stmt/3, execute_stmt/4]).
-export([disconnect/2, disconnect/3]).
-export([terminate_env/2, terminate_env/3]).

%% Start/stop
-export([start_link/2, start_link/3]).
-export([stop/1, stop/2]).

%% Miscellaneous
%-export([sql_func_exists/2]).
-export([display_size/2]).
%%------------------------------------------------------

%% Internal exports (for debug)
-export([get_state/1]).


%% Internal exports (for supervisor).
-export([start_link_sup/2, start_link_sup/3]).

%% gen_server callbacks
-export([init/1, handle_call/3, handle_cast/2, handle_info/2, terminate/2]).


%% Internal state
-record(state, {port,                        % The port to the c-node.
		connected = false,           % Connected to database. 
		clientPid,                   % Pid of the odbc-client.
		columnvector = []            % memoryref for columns
	       }).

%%---------------
%% Internal macros
%% Path to C server main
-define(ServerDir, filename:nativename(filename:join(code:priv_dir(odbc), "bin"))).

%% Name of C server main
-define(ServerProg, "odbcserver").

%% Default timeout.
-define(DefaultTimeout, 5000).


%%%----------------------------------------------
%% DEBUG MACRO

-ifdef (DEBUG).
-define(DBG(Format, Args), io:format("Debug Module:~p line:~p: "
				     ++Format++ "~n",
                                     [?MODULE,?LINE]++Args)).
-else.
-define(DBG(F,A),[]).
-endif.


%%%----------------------------------------------------------
%%% Start and Stop
%%%----------------------------------------------------------


%%---------------------------------------------------
%% start_link/[2, 3]
%%
%% Starts the ODBCE server and the C node.
%% Links to the server to the calling process.
%% Registers the new server with the supervisor.
%%
%% start_link(ServerName, Args, Options, Client) ->
%% start_link(Args, Options, Client) ->
%%    Result
%% Options -> [Opt]                ; List of options.
%% Opt -> {timeout, integer()} |   ; The time in ms allowed for initialisation,
%%                                   see gen_server.
%%        {debug, [Dbg]}             Debug options, see gen_server.
%% Dbg -> trace |                  ; Debug options, see gen_server.
%%        log |
%%        statistics |
%%        {log_to_file, FileName} |
%%        {install, {Func, FuncState}}
%% ServerName -> {local, atom()} |   ; When supplied, causes the server to be
%%               {global, atom()}      registered, locally or globally. If the
%%                                     server is started without a name it can
%%                                     only be called using the returned Pid.
%% Result -> {ok, pid()} |           ; The pid of the erl. server.
%%           {error, Reason}           Error tuple.
%% Reason -> {already_started, pid()}| ; Server already started.
%%           timeout |                 Timeout expired.
%%           {no_c_node, Info}         Can't start C node. The program may not
%%                                     have been found or may not have been
%%                                     executable e.g.
%% Info -> string()                  ; More information.
%%


start_link(Args, Options) ->
    chk_options(Options, start_link, [Args, Options], 2),
    supervisor:start_child(odbc_sup, [[{client, self()} | Args], Options]).

start_link(ServerName, Args, Options) ->
    chk_server(ServerName, start_link, [ServerName, Args, Options], 1),
    chk_options(Options, start_link, [ServerName, Args, Options], 3),
    supervisor:start_child(odbc_sup,
			   [ServerName, [{client, self()} | Args], Options]).



%% start_link_sup/[2, 3]
%%
%% Called by the supervisor to start a new server instance.
%% (start_link calls the supervisor which calls start_link_sup.)
%%
start_link_sup(Args, Options) ->
    gen_server:start_link(?MODULE, Args, Options).

start_link_sup(ServerName, Args, Options) ->
    gen_server:start_link(ServerName, ?MODULE, Args, Options).



%%-------------------------------------------------------
%% stop/[1, 2]
%%
%% Stops the ODBCE server and the C node.
%%
%% stop(Server) ->
%% stop(Server, Timeout) ->
%%    ok
%%
%% Server -> pid() |                 ; The pid of the server,
%%           Name |                    a registered name,
%%           {global, Name} |          a globally registered name, or
%%           {Name, Node}              a registered name on a remote node.
%% Timeout -> integer() |            ; Max time (ms) for serving the request.
%%            infinity
%%

stop(Server) ->
    stop(Server, ?DefaultTimeout).

stop(Server, Timeout) ->
    chk_timeout(Timeout, stop, [Server, Timeout], 2),
    gen_server:call(Server, stop, Timeout).




%%%--------------------------------------------------------------------------
%%% Basic API
%%%--------------------------------------------------------------------------

%% sqlBindColumn/[3, 4]
%%
%% Assigns a referens to a column
%% (binds a referens to a  column).
%%
%% sqlBindColumn(Server, ColNum, RefBuf) ->
%% sqlBindColumn(Server, ColNum, RefBuf, Timeout) ->
%%   Result
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% ColNum -> integer()                  ; Column number from left to right
%%                                        starting at 1.
%% RefBuf -> term()                     ; Reference to the buffer where the
%%                                        column data is placed (and to the
%%                                        associated length/indicator buffer).
%%                                        a buffer and a column.
%% Timeout -> integer() |               ; timeout (ms) for genserver
%%            infinity
%% Result -> ?SQL_SUCCESS |
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR
%%

sqlBindColumn(Server, ColNum, RefBuf) ->
    sqlBindColumn(Server, ColNum, RefBuf, ?DefaultTimeout).

sqlBindColumn(Server, ColNum, RefBuf, Timeout) ->
    CmdStr = "bincol",
    Ret = gen_server:call(Server,{bindcolumn, CmdStr, ColNum, RefBuf},
			  Timeout),
    ?DBG("sqlBindColumn gen_server return: ~p",[Ret]),
    case Ret of
	{ok, RetMsg, RetCode} ->
	    RetCode;
	{error, ErrMsg, ErrCode} ->
	    {error, ErrCode};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlCloseCursor/[1, 2]
%%
%% Closes a cursor that has been opened on a statement and discards pending 
%% results.
%%
%% sqlCloseCursor(Server) ->
%% sqlCloseCursor(Server, Timeout) ->
%%   Result
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% RefStmtHandle -> term()              ; Reference to the statement handle.
%% Timeout -> integer() |               ; timeout (ms) for genserver.
%%            infinity
%% Result -> ?SQL_SUCCESS |
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR

sqlCloseCursor(Server) ->
    sqlCloseCursor(Server, ?DefaultTimeout).

sqlCloseCursor(Server, Timeout) ->
    CmdStr = "closec",
    Ret = gen_server:call(Server,{closecursor, CmdStr},Timeout),
    ?DBG("sqlCloseCursor gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	{ok, Result, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlConnect/[4, 5]
%%
%% Establishes a connection to a driver and a data source.
%%
%% sqlConnect(Server, DSN, UID, Auth) ->
%% sqlConnect(Server, DSN, UID, Auth, Timeout) ->
%%   Result
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% DSN -> string()                      ; The name of the data source.
%% UID -> string()                      ; The user ID.
%% Auth -> string()                     ; The user's password.
%% Timeout -> integer() |               ; timeout (ms) for genserver.
%%            infinity
%% Result -> ?SQL_SUCCESS | 
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR

sqlConnect(Server, DSN, UID, Auth) ->
    sqlConnect(Server, DSN, UID, Auth, ?DefaultTimeout).

sqlConnect(Server, DSN, UID, Auth, Timeout) ->
    chk_str(DSN, sqlConnect,[Server, DSN, UID, Auth, Timeout],2),
    chk_str(UID, sqlConnect,[Server, DSN, UID, Auth, Timeout],3),
    chk_str(Auth, sqlConnect,[Server, DSN, UID, Auth, Timeout], 4),

%% When there are not any user passsword
    N = string:len(Auth),
    ConnectStr = case N of
	0 -> 
	    "opendb20DSN=" ++ DSN ++ ";UID=" ++ UID;
	NotZero-> 
	    "opendb20DSN=" ++ DSN ++ ";UID=" ++ UID ++ ";PWD="++ Auth
    end,
    Ret = gen_server:call(Server,{openconnection, ConnectStr, self()},
			  Timeout),
    ?DBG("sqlConnect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	{ok, Msg, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlDescribeCol/[2, 3]
%%
%% Returns the result descriptor -- column name, type, column size,
%% decimal digits, and nullability -- for one column in the result set.
%%
%% sqlDescribeCol(Server, ColNum) ->
%% sqlDescribeCol(Server, ColNum, Timeout) ->
%%   {Result, ColName, Nullable}
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% ColNum -> integer()                  ; Column number from left to right
%%                                        starting at 1.    
%% Timeout -> integer() |               ; timeout (ms) for genserver
%%            infinity
%% Result -> ?SQL_SUCCESS |
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR
%% ColName -> string()                  ; The column name.
%% Nullable -> ?SQL_NO_NULLS |          ; Indicates whether the column allows
%%             ?SQL_NULLABLE |            NULL values or not.
%%             ?SQL_NULLABLE_UNKNOWN
%%

sqlDescribeCol(Server, ColNum) ->
    sqlDescribeCol(Server, ColNum, ?DefaultTimeout).

sqlDescribeCol(Server, ColNum, Timeout) ->
    CmdStr = "descol" ++ integer_to_list(ColNum),
    Ret = gen_server:call(Server,{describeColumn, CmdStr},Timeout),
    ?DBG("sqlDescribeCol gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrCode};
	{ok, ReturnCode, CName, CNullable} ->
	    {ReturnCode, CName, CNullable};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlDisConnect/[1, 2]
%%
%% Close the connection.
%%
%% sqlDisConnect(Server) ->
%% sqlDisconnect(Server, Timeout) ->
%%   Result
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% Timeout -> integer() |               ; timeout (ms) for genserver.
%%            infinity
%% Result -> ?SQL_SUCCESS | 
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR
%% 

sqlDisConnect(Server) ->
    sqlDisConnect(Server, ?DefaultTimeout).

sqlDisConnect(Server, Timeout) ->
    CloseStr = "closdb",
    Ret = gen_server:call(Server,{closeconnection, CloseStr}, Timeout),
    ?DBG("sqlDisConnect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded disconnecting  
	{ok, Mes,RetCode} ->
	    RetCode;
	% Did not succeded disconnecting (may already be disconnected)
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	Other ->
	    {error, Other}
    end.

%% sqlExecDirect/[2, 3]
%%
%% Executes a preparable statement.
%%
%% sqlExecDirect(Server, Stmt) ->
%% sqlExecDirect(Server, Stmt, Timeout) ->
%%   Result
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% Stmt -> string()                     ; An SQL statement.
%% Timeout -> integer() |               ; timeout (ms) for genserver
%%            infinity
%% Result -> ?SQL_SUCCESS |
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_NEED_DATA |
%%           ?SQL_NO_DATA |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR

sqlExecDirect(Server, Stmt) ->
    sqlExecDirect(Server, Stmt, ?DefaultTimeout).

sqlExecDirect(Server, Stmt, Timeout) ->
    chk_str(Stmt, sqlExecuteDirect,[Server, Stmt, Timeout], 2),
    SqlStmt = "exedir20" ++ Stmt,
    Ret = gen_server:call(Server, {execdir,SqlStmt}, Timeout),
    ?DBG("sqlExecDirect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, Result, Errcode} ->
	    Errcode;
	{ok, Result, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlFetch/[1, 2]
%%
%% Fetches a row of data from a result set. The driver returns data for 
%% all columns that were bound to storage locations with sql_bind_col/[4, 5].
%%
%% sqlFetch(Server) ->
%% sqlFetch(Server, Timeout) ->
%%   Result
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% Timeout -> integer() |               ; timeout (ms) for genserver
%%            infinity
%% Result -> ?SQL_SUCCESS |  
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_NO_DATA |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR
%% 

sqlFetch(Server) ->
    sqlFetch(Server, ?DefaultTimeout).

sqlFetch(Server, Timeout) ->
    CmdStr = "fetchd",
    Ret = gen_server:call(Server,{fetch, CmdStr},Timeout),
    ?DBG("sqlFetct gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	{ok, ReturnCode} ->
	    ReturnCode;
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlNumResultCols/[1, 2]
%%
%% Returns the number of columns in a result set.
%%
%% sqlNumResultCols(Server) ->
%% sqlNumResultCols(Server, Timeout) ->
%%   {Result, ColCount}
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% Timeout -> integer() |               ; timeout (ms) for genserver
%%            infinity
%% Result -> ?SQL_SUCCESS |  
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR
%% ColCount -> integer()                ; number of columns in the result set.

sqlNumResultCols(Server) ->
    sqlNumResultCols(Server, ?DefaultTimeout).

sqlNumResultCols(Server, Timeout) ->
    CmdStr = "rescol",
    Ret = gen_server:call(Server,{numResultCols, CmdStr},Timeout),
    ?DBG("sqlNumResultCols gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    ErrCode;
	{ok, NumberColumn, ReturnCode} ->
	    {ReturnCode, NumberColumn};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%% sqlRowCount/[1, 2]
%%
%% Returns the number of rows affected by an UPDATE, INSERT, or DELETE 
%% statement.
%%
%% sqlRowCount(Server) ->
%% sqlRowCount(Server, Timeout) ->
%%   {Result, RowCount}
%%
%% Server -> pid() |                   ; The pid of the server,
%%           Name |                      a registered name,
%%           {global, Name} |            a globally registered name, or
%%           {Name, Node}                a registered name on a remote node.
%% Timeout -> integer() |              ; timeout (ms) for genserver.
%%            infinity
%% Result -> ?SQL_SUCCESS |
%%           ?SQL_SUCCESS_WITH_INFO |
%%           ?SQL_INVALID_HANDLE |
%%           ?SQL_ERROR
%% RowCount -> integer()               ; The number of affected rows (or -1
%%                                       if not available).
%%

sqlRowCount(Server) ->
    sqlRowCount(Server, ?DefaultTimeout).

sqlRowCount(Server, Timeout) ->
    CmdStr = "rowcon",
    Ret = gen_server:call(Server,{rowCount, CmdStr},Timeout),
    ?DBG("sqlRowCount gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in making ODBC call (which may have returned an error)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrCode};
	{ok, NumberColumn, ReturnCode} ->
	    {ReturnCode, NumberColumn};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.
    

%% columnRef/[0]
%%
%% Generate a referens for a column. 
%%
%% columnRef() ->
%%   {ok, columnRef}
%%
%% columnRef -> term()                   ; A referens to a column.
%%

columnRef() ->
    MemRef = erlang:make_ref(),
    {ok, MemRef}.


%% readData/[2, 3]
%%
%% Returns data from the SQL_fetch call.
%%
%% readData(Server, RefBuf) ->
%% readData(Server, RefBuf, Timeout) ->
%%   {ok, Value}
%%
%% Server -> pid() |                    ; The pid of the server,
%%           Name |                       a registered name,
%%           {global, Name} |             a globally registered name, or
%%           {Name, Node}                 a registered name on a remote node.
%% RefCol -> term()                     ; A referens to the column.
%% Timeout -> integer() |               ; timeout (ms) for genserver.
%%            infinity
%% Value -> string() |                  ; Returned Data 
%%

readData(Server, RefBuf) ->
    readData(Server, RefBuf, ?DefaultTimeout).

readData(Server, RefBuf, Timeout) ->
    CmdStr = "readbu",
    Ret = gen_server:call(Server, {readbuffer, CmdStr, RefBuf},Timeout),
    ?DBG("readData gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded in get data
	{ok, Data} ->
	    {ok, Data};
	% gen_server specific error
	{error, Reason} ->
	    exit(Reason);
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	Other ->
	    Other
    end.

%%%----------------------------------------------------------------------
%%% New Utility API
%%%
%%% All functionality of the Utility API is also available through
%%% use of functions of the Basic API.
%%%
%%%----------------------------------------------------------------------


%% erl_connect/[2, 3, 4, 5]
%%
%% Opens a connection to a data source. There can be only one open data 
%% source connection per server. erl_connect/[2, 3] is used when the 
%% information that can be supplied through erl_connect/[4, 5] does not 
%% suffice.
%% Equivalent to calling the following Basic API fcns:
%% - sql_alloc_connect
%% - sql_connect or sql_driver_connect
%% - sql_get_diag_rec
%%
%% erl_connect(Server, ConnectStr) ->
%% erl_connect(Server, ConnectStr, Timeout) ->
%% erl_connect(Server, DSN, UID, PWD) ->
%% erl_connect(Server, DSN, UID, PWD, Timeout) ->
%%   ok |
%%   {error, {Fcn, [Reason]}}
%%
%% Server -> pid() |                      ; The pid of the server,
%%           Name |                         a registered name,
%%           {global, Name} |               a globally registered name, or
%%           {Name, Node}                   a registered name on a remote node.
%% ConnectStr -> string()                 ; 
%% Timeout -> integer() |                 ; timeotu (ms) for genserver.
%%            infinity
%% DSN -> string()                        ; Name of the data source.
%% UID -> string()                        ; User ID.
%% PWD -> string()                        ; Password.
%% Fcn -> atom()                          ; The originating function.
%% Reason -> {Fcn, ErrMsg, Errcode}	  ; An ODBC error tuple.
%%

erl_connect(Server, ConnectStr) ->
    erl_connect(Server, ConnectStr, ?DefaultTimeout).

erl_connect(Server, ConnectStr, Timeout) ->
    chk_str(ConnectStr, erl_connect,[Server, ConnectStr, Timeout],2),
    NewConnectStr = "opendb20" ++ ConnectStr,
    Ret = gen_server:call(Server,{openconnection, NewConnectStr, self()}, 
			  Timeout),
    ?DBG("erl_connect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Server going down
	{stopped, Reason} ->
	    exit({stopped, Reason});
	% Succeeded connected 
	{ok, Msg, RetCode} ->
	    ok;
	% Did not succed connected
	{error, ErrMsg, ErrCode} ->
	    {error,ErrMsg, ErrCode};
	Other ->
	    Other
    end.

erl_connect(Server, DSN, UID, PWD) ->
    erl_connect(Server, DSN, UID, PWD, ?DefaultTimeout).

erl_connect(Server, DSN, UID, PWD, Timeout) ->
    chk_str(DSN, erl_connect,[Server, DSN, UID, PWD, Timeout], 2),
    chk_str(UID, erl_connect,[Server, DSN, UID, PWD, Timeout], 3),
    chk_str(PWD, erl_connect,[Server, DSN, UID, PWD, Timeout], 4),
    N = string:len(PWD),
    ConnectStr = case N of 
		     0 -> "DSN=" ++ DSN ++ ";UID=" ++ UID;
		     NotNull ->
			 "DSN=" ++ DSN ++ ";UID=" ++ UID ++ ";PWD="++ PWD
		 end,
    erl_connect(Server, ConnectStr, Timeout).


%% erl_executeStmt/[2, 3]
%%
%% Executes a single SQL statement. All changes to the data source are
%% automatically committed if successful.
%% Equivalent to calling the following sql-fcns:
%% - sql_alloc_stmt
%% - sql_exec_direct
%% - sql_num_result_cols
%% - sql_row_count
%% - sql_end_tran
%% - sql_describe_col
%% - alloc_buffer
%% - sql_bind_col
%% - sql_fetch
%% - read_buffer
%% - sql_free_handle
%% - dealloc_buffer
%% - sql_get_diag_rec
%%
%% erl_executeStmt(Server, Stmt) ->
%% erl_executeStmt(Server, Stmt, Timeout) ->
%%   {updated, NRows} |
%%   {selected, [ColName], [Row]}
%%   {error, {Fcn, [Reason]}}
%%
%% Server -> pid() |                     ; The pid of the server,
%%           Name |                        a registered name,
%%           {global, Name} |              a globally registered name, or
%%           {Name, Node}                  a registered name on a remote node.
%% Stmt -> string()                      ; SQL statement to execute.
%% Timeout -> integer() |                ; timeout (ms) for genserver.
%%            infinity
%% NRows -> integer()                    ; The number of updated rows.
%% ColName -> string()                   ; The name of the columns in the table.
%% Row -> [Value]                        ; One row of the resulting table.
%%   Value -> string                     ; One value in a row.
%% Fcn -> atom()                         ; The originating function.
%% Reason -> {Fcn, SqlState, MoreInfo}	 ; An ODBC error tuple.
%%   SqlState -> string()                   ; The SQL state
%%   MoreInfo -> {NativeCode, Msg, LenMsg}  ; More error info.
%%     NativeCode -> string()               ; Data source specific error code.
%%     Msg -> string()                      ; Error message.
%%     LenMsg -> integer()                  ; Length of Msg.
%%

erl_executeStmt(Server, Stmt) ->
    erl_executeStmt(Server, Stmt, ?DefaultTimeout).

erl_executeStmt(Server, Stmt, Timeout) ->
    chk_str(Stmt, execute_stmt,[Server, Stmt, Timeout], 2),
    SqlStmt = "execdb20" ++ Stmt,
    case gen_server:call(Server,{execute, SqlStmt}, Timeout)   of
	ok -> wait_result([],[]);
	    
	% Succeeded in making ODBC call (which may have returned an error)
	Other ->
	    Other
    end.
wait_result(ColumnNameList,RowList) ->
    receive
	{column_name,ColumnName} ->
	   ?DBG("erl_executeStmt -> ColumnName: ~p~n",[ColumnName]),
	   wait_result(ColumnName,RowList);
	{row,Data} ->
	    ?DBG("erl_executeStmt -> {row, Data}: ~p",[Data]),
	    wait_result(ColumnNameList,[Data|RowList]);
	{selected, lastrow} ->
	    RevL = lists:reverse(RowList),
	    {selected,ColumnNameList ,RevL};
	{updated, NupdatRow} ->
	    {updated, NupdatRow};
	{error, ErrMsg} ->
	    {error,ErrMsg};
	Other ->
	    ?DBG("erl_executeStmt -> Other: ~p~n",[Other]),
	    Other
    end.


%% erl_disconnect/[2, 3]
%%
%% Closes the connection to a data source.
%% erl_disconnect(Server, RefConnHandle) ->
%% erl_disconnect(Server, RefConnHandle, Timeout) ->
%%   ok |
%%   {error, {Fcn, [Reason]}}
%%
%% Server -> pid() |                     ; The pid of the server,
%%           Name |                        a registered name,
%%           {global, Name} |              a globally registered name, or
%%           {Name, Node}                  a registered name on a remote node.
%%
%% Timeout -> integer() |                ; timeout (ms) for genserver.
%%            infinity
%% Fcn -> atom()                         ; The originating function.
%% Reason -> {Fcn, ErrMsg, ErrCode}	 ; An ODBC error tuple.
%%   ErrMsg -> string()                  ; Error messages
%%   ErrCode -> integer()                ; SQL errorcode
%%

erl_disconnect(Server, RefConnHandle) ->
    erl_disconnect(Server, RefConnHandle, ?DefaultTimeout).

erl_disconnect(Server, RefConnHandle, Timeout) ->
    CloseStr = "closdb",
    Ret = gen_server:call(Server,{closeconnection, CloseStr}, Timeout),
    ?DBG("erl_disconnect gen_server call returns: ~p",[Ret]),
    case Ret of
	% Succeeded disconnecting  
	{ok, Mes,RetCode} ->
	    ok;
	% Did not succeded disconnecting (may already be disconnected)
	{error, ErrMsg, ErrCode} ->
	    {error, ErrMsg, ErrCode};

	Other ->
	    {error, Other}
    end.



%% This function is deprecated and is included only for compatibility reason.
%%--------------------------------------------------------------------------- 
sql_alloc_handle(Server, HandleType, RefInputHandle) ->
    sql_alloc_handle(Server, HandleType, RefInputHandle, ?DefaultTimeout).

sql_alloc_handle(Server, HandleType, RefInputHandle, Timeout) ->
    {0,void}. 

%% sql_bind_col/[4, 5]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_bind_col(Server, RefStmtHandle, ColNum, RefBuf) ->
    sql_bind_col(Server, RefStmtHandle, ColNum, RefBuf, ?DefaultTimeout).

sql_bind_col(Server, RefStmtHandle, ColNum, RefBuf, Timeout) ->
    chk_timeout(Timeout,
		sql_bind_col,
		[Server,  RefStmtHandle, Timeout],5),
    Ret = sqlBindColumn(Server, ColNum, RefBuf, Timeout),
    case Ret of
	{error, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% sql_close_cursor/[2, 3]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_close_cursor(Server, RefStmtHandle) ->
    sql_close_cursor(Server, RefStmtHandle, ?DefaultTimeout).

sql_close_cursor(Server, RefStmtHandle, Timeout) ->
    chk_timeout(Timeout,
		sql_close_cursor,
		[Server,  RefStmtHandle, Timeout],3),
    sqlCloseCursor(Server, Timeout).


%% sql_connect/[5, 6]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_connect(Server, RefConnHandle, DSN, UID, Auth) ->
    sql_connect(Server, RefConnHandle, DSN, UID, Auth, ?DefaultTimeout).

sql_connect(Server, RefConnHandle, DSN, UID, Auth, Timeout) ->
    chk_str(DSN,
	    sql_connect,
	    [Server, RefConnHandle, DSN, UID, Auth, Timeout],3),
    chk_str(UID,
	    sql_connect,
	    [Server, RefConnHandle, DSN, UID, Auth, Timeout],4),
    chk_str(Auth,
	    sql_connect,
	    [Server, RefConnHandle, DSN, UID, Auth, Timeout],5),
    chk_timeout(Timeout,
		sql_connect,
		[Server, RefConnHandle, DSN, UID, Auth, Timeout],6),
    sqlConnect(Server, DSN, UID, Auth, Timeout).

%% sql_describe_col/[4, 5]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_describe_col(Server, RefStmtHandle, ColNum, BufLenColName) ->
    sql_describe_col(Server, RefStmtHandle, ColNum, BufLenColName, ?DefaultTimeout).

sql_describe_col(Server, RefStmtHandle, ColNum, BufLenColName, Timeout) ->
    chk_timeout(Timeout,
		sql_describe_col,
		[Server,  RefStmtHandle, ColNum, BufLenColName, Timeout],6),
    Ret = sqlDescribeCol(Server, ColNum, Timeout),
    case Ret of
	{ReturnCode, ColName, CNullable} ->
	    {ReturnCode, {ColName, -1}, -1, -1, -1, CNullable};
	{error, ErrCode} ->
	    ErrCode;
	Other ->
	    Other
    end.

%% sql_disconnect/[2, 3]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_disconnect(Server, RefConnHandle) ->
    sql_disconnect(Server, RefConnHandle, ?DefaultTimeout).

sql_disconnect(Server, RefConnHandle, Timeout) ->
    chk_timeout(Timeout, sql_disconnect, [Server, RefConnHandle, Timeout], 3),
    sqlDisConnect(Server, Timeout).

%% This function is deprecated and is included only for compatibility reason.
%%--------------------------------------------------------------------------
sql_driver_connect(Server, RefConnHandle, InConnStr,
		   BufLenOutConnStr, DrvCompletion) ->
    sql_driver_connect(Server, RefConnHandle, InConnStr,
		       BufLenOutConnStr, DrvCompletion, ?DefaultTimeout).

sql_driver_connect(Server, RefConnHandle, InConnStr,
		   BufLenOutConnStr, DrvCompletion, Timeout) ->
    {0, void}.


%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_end_tran(Server, HandleType, RefHandle, ComplType) ->
    sql_end_tran(Server, HandleType, RefHandle, ComplType, ?DefaultTimeout).

sql_end_tran(Server, HandleType, RefHandle, ComplType, Timeout) ->
    0.

%% sql_exec_direct/[3, 4]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_exec_direct(Server, RefStmtHandle, Stmt) ->
    sql_exec_direct(Server, RefStmtHandle, Stmt, ?DefaultTimeout).


sql_exec_direct(Server, RefStmtHandle, Stmt, Timeout) ->
    chk_str(Stmt, sql_exec_direct, [Server, RefStmtHandle, Stmt, Timeout], 3),
    chk_timeout(Timeout, sql_exec_direct, [Server, RefStmtHandle, Stmt, 
					   Timeout], 4),
    sqlExecDirect(Server, Stmt, Timeout).

%% sql_fetch/[2, 3]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_fetch(Server, RefStmtHandle) ->
    sql_fetch(Server, RefStmtHandle, ?DefaultTimeout).

sql_fetch(Server, RefStmtHandle, Timeout) ->
    chk_timeout(Timeout, sql_fetch, [Server, RefStmtHandle, Timeout], 3),
    sqlFetch(Server, Timeout).


%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_free_handle(Server, HandleType, RefHandle) ->
    sql_free_handle(Server, HandleType, RefHandle, ?DefaultTimeout).

sql_free_handle(Server, HandleType, RefHandle, Timeout) ->
    0.

%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_get_connect_attr(Server, RefConnHandle, Attr, BufType) ->
    sql_get_connect_attr(Server, RefConnHandle, Attr, BufType, 
			 ?DefaultTimeout).

sql_get_connect_attr(Server, RefConnHandle, Attr, BufType, Timeout) ->
    {0, -1}.

%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_get_diag_rec(Server, HandleType, RefHandle, RecNum, BufLenErrMsg) ->
    sql_get_diag_rec(Server, HandleType, RefHandle, RecNum, BufLenErrMsg,
		     ?DefaultTimeout).

sql_get_diag_rec(Server, HandleType, RefHandle, RecNum, BufLenErrMsg, 
		 Timeout) ->
   0.

%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_num_result_cols(Server, RefStmtHandle) ->
    sql_num_result_cols(Server, RefStmtHandle, ?DefaultTimeout).

sql_num_result_cols(Server, RefStmtHandle, Timeout) ->
    chk_timeout(Timeout, sql_num_result_cols, [Server, RefStmtHandle, 
				   Timeout], 3),
    sqlNumResultCols(Server, Timeout).

%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
sql_row_count(Server, RefStmtHandle) ->
    sql_row_count(Server, RefStmtHandle, ?DefaultTimeout).

sql_row_count(Server, RefStmtHandle, Timeout) ->
    chk_timeout(Timeout, sql_row_count, [Server, RefStmtHandle, 
				   Timeout], 3),
    Ret = sqlRowCount(Server, Timeout),
    case Ret of
	{error, ErrCode} ->
	    ErrCode;
	 Other ->
	    Other
    end.

%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_set_connect_attr(Server, RefConnHandle, Attr, Value, BufType) ->
    sql_set_connect_attr(Server, RefConnHandle, Attr, Value, BufType,
			 ?DefaultTimeout).

sql_set_connect_attr(Server, RefConnHandle, Attr, Value, BufType, Timeout) ->
    0.

%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
sql_set_env_attr(Server, RefEnvHandle, Attr, Value, BufType) ->
    sql_set_env_attr(Server, RefEnvHandle, Attr, Value, BufType, 
		     ?DefaultTimeout).

sql_set_env_attr(Server, RefEnvHandle, Attr, Value, BufType, Timeout) ->
    0.

%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
alloc_buffer(Server, BufCType, Size) ->
    alloc_buffer(Server, BufCType, Size, ?DefaultTimeout).

alloc_buffer(Server, BufCType, Size, Timeout) ->
    columnRef().

%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
dealloc_buffer(Server, RefBuf) ->
    dealloc_buffer(Server, RefBuf, ?DefaultTimeout).

dealloc_buffer(Server, RefBuf, Timeout) ->
    ok.

%% read_buffer/[2, 3]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
read_buffer(Server, RefBuf) ->
    read_buffer(Server, RefBuf, ?DefaultTimeout).

read_buffer(Server, RefBuf, Timeout) ->
    chk_timeout(Timeout, read_buffer, [Server, RefBuf, Timeout], 3),
    Ret = readData(Server, RefBuf, Timeout),
    case Ret of
	{ok, Value} ->
	    {ok, {Value, -1}}; 
	Other ->
	    Other
    end.

%%%----------------------------------------------------------------------
%%% Utility API
%%%
%%% All functionality of the Utility API is also available through
%%% use of functions of the Basic API.
%%%
%%%----------------------------------------------------------------------

%% This function is deprecated and is included only for compatibility reason.
%%---------------------------------------------------------------------------
init_env(Server) ->
    init_env(Server, ?DefaultTimeout).

init_env(Server, Timeout) ->
    {ok, void}.



%% connect/[3, 4, 5, 6]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
connect(Server, RefEnvHandle, ConnectStr) ->
    connect(Server, RefEnvHandle, ConnectStr, ?DefaultTimeout).

connect(Server, RefEnvHandle, ConnectStr, Timeout) ->
    chk_str(ConnectStr, connect, [Server, RefEnvHandle, ConnectStr, 
				  Timeout], 3),
    chk_timeout(Timeout, connect, [Server, RefEnvHandle, ConnectStr, 
				   Timeout], 4),
    Ret = erl_connect(Server, ConnectStr, Timeout),
    case Ret of
	ok ->
	    {ok, void};
	{error,ErrMsg, ErrCode} ->
	    {error, {connect, [{void, {ErrCode, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.

connect(Server, RefEnvHandle, DSN, UID, PWD) ->
    connect(Server, RefEnvHandle, DSN, UID, PWD, ?DefaultTimeout).

connect(Server, RefEnvHandle, DSN, UID, PWD, Timeout) ->
    chk_str(DSN, connect, [Server, RefEnvHandle, DSN, UID, PWD, Timeout], 3),
    chk_str(UID, connect, [Server, RefEnvHandle, DSN, UID, PWD, Timeout], 4),
    chk_str(PWD, connect, [Server, RefEnvHandle, DSN, UID, PWD, Timeout], 5),
    Ret = erl_connect(Server, DSN, UID, PWD, Timeout),
    case Ret of
	ok ->
	    {ok, void};
	{error,ErrMsg, ErrCode} ->
	    {error, {connect, [{void, {ErrCode, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.

%% execute_stmt/[3, 4]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
execute_stmt(Server, RefConnHandle, Stmt) ->
    execute_stmt(Server, RefConnHandle, Stmt, ?DefaultTimeout).

execute_stmt(Server, RefConnHandle, Stmt, Timeout) ->
    chk_str(Stmt, execute_stmt, [Server, RefConnHandle, Stmt, Timeout], 3),
    chk_timeout(Timeout, execute_stmt, [Server, RefConnHandle, Stmt, 
					Timeout], 4),
    Ret = erl_executeStmt(Server, Stmt, Timeout),
    case Ret of
	{error,ErrMsg} ->
	    {error,{execute_stmt, [{void, {-1, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.


%% disconnect/[2, 3]
%% This function is deprecated and is included only for compatibility reason.
%% This function calls a new function in ODBC.
%%---------------------------------------------------------------------------
disconnect(Server, RefConnHandle) ->
    disconnect(Server, RefConnHandle, ?DefaultTimeout).

disconnect(Server, RefConnHandle, Timeout) ->
    chk_timeout(Timeout, disconnect, [Server, RefConnHandle, Timeout], 3),
    Ret = erl_disconnect(Server, Timeout),
    case Ret of
	ok ->
	    ok;
	{error,ErrMsg, ErrCode} ->
	    {error, {disconnect, [{void, {ErrCode, ErrMsg, -1}}]}};
	Other ->
	    Other
    end.

%% This function is deprecated and is included only for compatibility reason.
%%-------------------------------------------------------------------
terminate_env(Server, RefEnvHandle) ->
    terminate_env(Server, RefEnvHandle, ?DefaultTimeout).

terminate_env(Server, RefEnvHandle, Timeout) ->
    ok.

%% This function is deprecated and is included only for compatibility reason.
%%-------------------------------------------------------------------
display_size(SqlType, ColumnSize) when 
  integer(SqlType), integer(ColumnSize) ->
    -1.


%%%----------------------------------------------------------------------
%%% Debug fcns
%%%----------------------------------------------------------------------

%% Returns the internal state in a tagged tuple list.
%%
get_state(Server) ->
    State = gen_server:call(Server, get_state),
    [{port, State#state.port},
    {connected, State#state.connected},
    {clientPid, State#state.clientPid},
    {columnvector, State#state.columnvector}].


%%%======================================================================
%%% Callback functions from gen_server
%%%======================================================================

%%----------------------------------------------------------------------
%% Func: init/1
%% Returns: {ok, State}          |
%%          {ok, State, Timeout} |
%%          {stop, Reason}
%%----------------------------------------------------------------------

%%
init(Args) ->
    process_flag(trap_exit, true),
    % Start the C node
    case os:find_executable(?ServerProg, ?ServerDir) of
	FileName when list(FileName)->
	    Port = open_port({spawn,FileName},[{packet,4}, binary]),
	    State = #state{port=Port},
	    {ok, State};
	false ->
	    {stop, {no_c_node, "Can't find executable."}}
    end.
		    

%%----------------------------------------------------------------------
%% Func: handle_call/3
%% Returns: {reply, Reply, State}          |
%%          {reply, Reply, State, Timeout} |
%%          {noreply, State}               |
%%          {noreply, State, Timeout}      |
%%          {stop, Reason, Reply, State}   | (terminate/2 is called)
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------

%%---------------------------------------------------------------------
%% Stop Call Backs
%%---------------------------------------------------------------------

% stop
handle_call(stop, _From, State) ->
    {stop, normal, ok, State};

%%---------------------------------------------------------------------
%% ODBC Call Backs
%%---------------------------------------------------------------------

handle_call({openconnection,ConnectString,ClientPid}, _, State) ->
    Port = State#state.port,
    case send(Port, ConnectString) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    NewState = State#state{connected = true, clientPid = ClientPid},
	    {reply, Reply, NewState}
    end;
handle_call({execute, SqlStmt}, _, State) ->
    Port = State#state.port,
    Connected = State#state.connected,
    Reply = case Connected of
		true ->
		    Port ! {self(), {command, term_to_binary(SqlStmt)}},
		    ok;
		false ->
		    {error, not_connected}
	    end,
    {reply, Reply, State};
handle_call({execdir, CmdStr}, _, State) ->
    Port = State#state.port,
    case send(Port, CmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    {reply,Reply, State}
    end;
handle_call({closeconnection,CloseStr}, _,State) ->
    Port = State#state.port,
    case send(Port, CloseStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    NewState = State#state{connected = false},
	    {reply,Reply, NewState}
    end;
handle_call({bindcolumn, CmdStr, ColNum, MemRef}, _,State) ->
    ColumnVec = State#state.columnvector,
    Port = State#state.port,
    NewColumnVec = insert({MemRef,ColNum}, ColumnVec),
    ?DBG("handle_call: bindcolumn insert return ~p",[NewColumnVec]),
    NewState = State#state{columnvector = NewColumnVec},
    NCmdStr = CmdStr ++ integer_to_list(ColNum),
    case send(Port, NCmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, NewState};
	Reply ->
	    {reply,Reply, NewState}
    end;
handle_call({describeColumn, CmdStr}, _, State) ->
    Port = State#state.port,
    case send(Port, CmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    {reply,Reply, State}
    end;
handle_call({numResultCols, CmdStr}, _, State) ->
    Port = State#state.port,
    case send(Port, CmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    {reply,Reply, State}
    end;
handle_call({rowCount, CmdStr}, _, State) ->
    Port = State#state.port,
    case send(Port, CmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    {reply,Reply, State}
    end;
handle_call({fetch, CmdStr}, _, State) ->
    Port = State#state.port,
    case send(Port, CmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    {reply,Reply, State}
    end;
handle_call({readbuffer, CmdStr, ColumnRef}, _, State) ->
    Port = State#state.port,
    ColumnVect = State#state.columnvector,
    ColumnNumber = lockup(ColumnRef,ColumnVect),
    ?DBG("handle_call: readbuffer lockup return ~p",[ColumnNumber]),
    NewCmdStr = CmdStr ++ integer_to_list(ColumnNumber), 
    case send(Port, NewCmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, State};
	Reply ->
	    {reply,Reply, State}
    end;
handle_call({closecursor, CmdStr}, _, State) ->
    Port = State#state.port,
    ColumnVect = State#state.columnvector,
    NewColumnVect = delete(ColumnVect),
    NewState = State#state{columnvector = NewColumnVect},
    case send(Port, CmdStr) of
	{error, Reason} ->
	    {stop, lost_connection, NewState};
	Reply ->
	    {reply,Reply, NewState}
    end;
handle_call(stop, _, State) ->
    ?DBG("handle_call(stop)~n",[]),
    {stop,  normal, ok, State};

%%---------------------------------------------------------------------
%% Debug Call Back
%%---------------------------------------------------------------------

handle_call(get_state, _From, State) ->
    {reply, State, State};


%%-----------------------------------------------------------

% Unknown request
handle_call(Request, _From, State) ->
    Reply = ok,
    {reply, Reply, State}.
%%----------------------------------------------------------------------


%%----------------------------------------------------------------------
%% Func: handle_cast/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------
handle_cast(_Msg, State) ->
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: handle_info/2
%% Returns: {noreply, State}          |
%%          {noreply, State, Timeout} |
%%          {stop, Reason, State}            (terminate/2 is called)
%%----------------------------------------------------------------------

handle_info({'EXIT', Pid, Reason}, State) when pid(Pid) ->
    ok =
	error_logger:error_report({"ODBC: Linked process:", 
				   {'EXIT', Pid, Reason}}),
    {stop, {stopped, {'EXIT', Pid, Reason}}, State};
% EXIT from port.
% Log exit and terminate.
%
handle_info({'EXIT', Port, Reason}, State) ->
    ok =
	error_logger:error_report({"ODBC: Port:", {'EXIT', Port, Reason}}),
    {stop, {stopped, {'EXIT', Port, Reason}}, State};

handle_info({Port, {data, BinData}}, State) ->
    ClientPid = State#state.clientPid,
    case (catch binary_to_term(BinData)) of
	{'EXIT',Reason} ->
	    ?DBG("handle_info(Port) decode error: ~p~n", [Reason]),
	    {error, Reason};
	Data ->
	    ?DBG("handle_info(Port) Data: ~p~n",[Data]),
	    ClientPid ! Data,
	    Data
    end,
    {noreply, State};

% Other.
%
handle_info(_Info, State) ->
    {noreply, State}.

%%----------------------------------------------------------------------
%% Func: terminate/2
%% Purpose: Shutdown the server
%% Returns: any (ignored by gen_server)
%%----------------------------------------------------------------------
terminate(Reason, State) ->
    Port = State#state.port,
    Port ! {self(), close}.



%%%======================================================================
%%% Internal functions
%%%======================================================================


%%%---------------------------------------------------------------------
%%% Argument checking
%%%---------------------------------------------------------------------
%%% These fcns check input arguments to the interface
%%% fcns. They all return ok. If an argument is not
%%% valid, the following exception thrown:
%%% {badarg, M, F, A, ArgNo, Info}
%%%   M -> atom()             ; Module.
%%%   F -> atom()             ; Function.
%%%   A -> [term()]           ; Argument list.
%%%   ArgNo -> integer()      ; The ordinal number of the argument in the
%%%                             interface fcn.
%%%   Info -> term()          ; Error info.
%%%
%%%---------------------------------------------------------------------



%%--------------------------------------------------------
%% chk_options/4
%%
%% Checks that Options is a list of supported options.
%%
%% chk_options(Options, CallingFunc, OrigArgs, ArgNo) -> ok
%%

% done checking
chk_options([], _CallingFunc, _OrigArgs, _ArgNo) ->
    ok;

% first option is {timeout, Value}
chk_options([{timeout, T} | Opts], CallingFunc, OrigArgs, ArgNo) when integer(T) ->
    chk_options(Opts, CallingFunc, OrigArgs, ArgNo);
chk_options([{timeout, T} | Opts], CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", {timeout, T}}});

% first option is {debug, Dbg}
chk_options([{debug, Dbg} | Opts], CallingFunc, OrigArgs, ArgNo) when list(Dbg) ->
    chk_dbg(Dbg, CallingFunc, OrigArgs, ArgNo),
    chk_options(Opts, CallingFunc, OrigArgs, ArgNo);
chk_options([{debug, Dbg} | Opts], CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", {debug, Dbg}}});

% first option is unsupported
chk_options([Opt | _Opts], CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Unknown option", Opt}});

% Opts not even a list
chk_options(Opts, CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", Opts}}).


% check debug options as far as reasonable
chk_dbg([], _CallingFunc, _OrigArgs, _ArgNo) ->
    ok;
chk_dbg([trace | Dbgs], CallingFunc, OrigArgs, ArgNo) ->
    chk_dbg(Dbgs, CallingFunc, OrigArgs, ArgNo);
chk_dbg([log | Dbgs], CallingFunc, OrigArgs, ArgNo) ->
    chk_dbg(Dbgs, CallingFunc, OrigArgs, ArgNo);
chk_dbg([statistics | Dbgs], CallingFunc, OrigArgs, ArgNo) ->
    chk_dbg(Dbgs, CallingFunc, OrigArgs, ArgNo);
chk_dbg([{log_to_file, FileName} | Dbgs], CallingFunc, OrigArgs, ArgNo)
  when list(FileName) ->
    chk_dbg(Dbgs, CallingFunc, OrigArgs, ArgNo);
chk_dbg([{log_to_file, FileName} | Dbgs], CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", {log_to_file, FileName}}});
chk_dbg([{install, {Func, _FuncState}} | Dbgs], CallingFunc, OrigArgs, ArgNo)
  when function(Func) ->
    chk_dbg(Dbgs, CallingFunc, OrigArgs, ArgNo);
chk_dbg([{install, {Func, FuncState}} | Dbgs], CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", {install, {Func, FuncState}}}});
chk_dbg([Dbg | _Dbgs], CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Unknown debug option", Dbg}}).


%%------------------------------------------------------------
%% chk_server/4
%%
%% Checks that Server is a type correct server name
%% (according to gen_server spec.)
%%
%% chk_server(Server, CallingFunc, OrigArgs, ArgNo) -> ok
%%
chk_server({local, Name}, _CallingFunc, _OrigArgs, _ArgNo) when atom(Name) ->
    ok;
chk_server({global, Name}, _CallingFunc, _OrigArgs, _ArgNo) when atom(Name) ->
    ok;
chk_server(Server, CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", Server}}).

    

%%------------------------------------------------------------------
%% chk_timeout/4
%%
%% Checks that Timeout is a valid value.
%%
%% chk_timeout(Timeout, CallingFunc, OrigArgs, ArgNo) -> ok
%%
chk_timeout(Timeout, _CallingFunc, _OrigArgs, _ArgNo) when integer(Timeout) ->
    ok;
chk_timeout(infinity, _CallingFunc, _OrigArgs, _ArgNo) ->
    ok;
chk_timeout(Timeout, CallingFunc, OrigArgs, ArgNo) when atom(Timeout) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Illegal value", Timeout}});
chk_timeout(Timeout, CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", Timeout}}).


%%------------------------------------------------------------------
%% chk_str/4
%%
%% Checks that a string is valid.
%% NOTE: Does not check that the string contains ASCII
%%       characters only or that it doesn't contain a
%%       null-termination character!
%%
%% chk_str(Str, CallingFunc, OrigArgs, ArgNo) -> ok
%%

chk_str(Str, _CallingFunc, _OrigArgs, _ArgNo) when list(Str) ->
    ok;
chk_str(Str, CallingFunc, OrigArgs, ArgNo) ->
    exit({badarg,
	  ?MODULE,
	  CallingFunc,
	  OrigArgs,
	  ArgNo,
	  {"Wrong type", Str}}).


%%%----------------------------------------------------------------
%%% Other Internal Fcns
%%%----------------------------------------------------------------


%%----------------------------------------------------------------
%% get_port_reply/1
%%
%% get_port_reply(Port) -> ok | {error, Reason}
%%
%% Returns ok if the tuple {eol, "running"} is received from
%% Port. Returns {error, Reason} if something else is received
%% from Port. Hangs if nothing is received.
%%

%% Send a command to ODBC
send(Port, Data)->
    Port ! {self(), {command, term_to_binary(Data)}},
    get_port_reply(Port).
%% Wait of respond from odbc
get_port_reply(Port)->
    receive
	{Port, {data, BinData}}  -> 
	    case (catch binary_to_term(BinData)) of
		{'EXIT',Reason} ->
		    {error, binary_to_term};
		Data ->
		    Data
	    end;
	{'EXIT', P, Reason} ->
	    {error, port_died};
	Other ->
	    ?DBG("get_prot_reply Other ~p ~n", [Other])
    end.

%% This is a simple table. The table contains tuples.
%% Each tuple contain a memoryreferens and a corrsponding
%% columnnumber.

insert({Ref, ColumnNr}, []) ->
    [{Ref, ColumnNr}];
insert({Ref, ColumnNr}, Xs) ->
    append(Xs,[{Ref, ColumnNr}]).

lockup(Ref, []) ->
    not_found;
lockup(Ref, [{Ref, ColumnNr} | Xs]) ->
    ColumnNr;
lockup(Ref, [{ORef, ColumnNr} | Xs] ) ->
    lockup(Ref, Xs).

delete(Xs) ->
    [].

append([], Ys) ->
    Ys;
append([X|Xs], Ys) ->
    [X|append(Xs, Ys)].


















