%% =====================================================================
%% Support functions for property lists
%%
%% Property lists are ordinary lists containing entries in the form of
%% tuples whose first elements are keys used for lookup and insertion,
%% together with atoms, which work as shorthand for tuples `{Atom,
%% true}'. (Terms not matching this description are allowed as elements
%% in the lists, but are ignored by this module.) If there is more than
%% one entry in a list for a certain key, the first occurrence overrides
%% any later (irrespective of the arity of the tuples). Property lists
%% are useful for representing inherited properties, such as options
%% passed to a function where a user may specify options overriding the
%% default settings, object properties, annotations, etc.
%%
%% Copyright (C) 2000 Richard Carlsson
%%
%% $Id: property_lists.erl,v 1.14 2001/05/07 15:08:14 richardc Exp $
%% ---------------------------------------------------------------------
%% This library is free software; you can redistribute it and/or modify
%% it under the terms of the GNU Lesser General Public License as
%% published by the Free Software Foundation; either version 2 of the
%% License, or (at your option) any later version.
%%
%% This library is distributed in the hope that it will be useful, but
%% WITHOUT ANY WARRANTY; without even the implied warranty of
%% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
%% Lesser General Public License for more details.
%%
%% You should have received a copy of the GNU Lesser General Public
%% License along with this library; if not, write to the Free Software
%% Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
%% USA
%%
%% Author contact: richardc@csd.uu.se
%% =====================================================================

-module(property_lists).

-export([property/1, property/2, lookup/2, is_defined/2, get_value/2,
	 get_value/3, get_bool/2, enter/2, enter/3, delete/2, concat/2,
	 compact/1, unfold/1, map/2, filter/2, get_keys/1,
	 substitute_aliases/2, substitute_negations/2, expand/2,
	 normalize/4]).


%% =====================================================================
%% property(Property) -> Property1
%% property(Key, Value) -> Property
%%
%%	    Property = Property1 = Name | {Key, ...} | term()
%%	    Name = atom()
%%	    Key = Name | term()
%%	    Value = term()
%%
%%	These are simple utility functions for creating a minimal
%%	(normal form) representation of a property:
%%
%%	`property(Property)' returns `Key' if `Property' is `{Key,
%%	true}' and `Key' is an atom; otherwise it returns `Property'.
%%
%%	`property(Key, Value)' returns `Key' if `Value' is the atom
%%	`true' and `Key' is an atom; otherwise it returns `{Key,
%%	Value}'.

property({Key, true}) when atom(Key) ->
    Key;
property(Property) ->
    Property.

property(Key, true) when atom(Key) ->
    Key;
property(Key, Value) ->
    {Key, Value}.


%% =====================================================================
%% lookup(Key, List) -> none | {Key, ...}
%%
%%	    Key = Name | term()
%%	    Name = atom()
%%	    List = [Name | {Key, ...} | term()]
%%
%%	Returns the first entry associated with the given key `Key' in
%%	the property list `List', if one exists, or otherwise `none'.
%%	For a tuple T = `{Key, ...}' in the list, T itself is the entry
%%	associated with `Key'. For an atom A in the list, the tuple `{A,
%%	true}' is the entry associated with `Key' if A = `Key'.

lookup(Key, [P | Ps]) ->
    if atom(P), P =:= Key ->
	    {Key, true};
       tuple(P), element(1, P) =:= Key ->
	    %% Note that if `P' is a 0-tuple, the guard will fail
	    %% to the next clause, not crash. Also note that `Key'
	    %% does not have to be an atom in this case.
	    P;
       true ->
	    lookup(Key, Ps)
    end;
lookup(Key, []) ->
    none.


%% =====================================================================
%% is_defined(Key, List) -> bool()
%%
%%	    Key = Name | term()
%%	    Name = atom()
%%	    List = [Name | {Key, ...} | term()]
%%
%%	Returns `true' if `List' contains an element `{Key, ...}', or if
%%	`Key' is an atom, an element `Key' or `{Key, ...}', otherwise
%%	returns `false'.

is_defined(Key, [P | Ps]) ->
    if atom(P), P =:= Key ->
	    true;
       tuple(P), element(1, P) =:= Key ->
	    true;
       true ->
	    is_defined(Key, Ps)
    end;
is_defined(Key, []) ->
    false.


%% =====================================================================
%% get_value(Key, List) -> term()
%% get_value(Key, List, Default) -> term()
%%
%%	    Key = Name | term()
%%	    Name = atom()
%%	    List = [Name | {Key, ...} | term()]
%%
%%	If `lookup(Key, List)' would yield `{Key, Value}', this function
%%	returns the corresponding `Value'; otherwise the value `Default'
%%	is returned. Thus, the entry associated with `Key' is assumed to
%%	be a 2-tuple, and if it is not, or if there is no entry
%%	associated with `Key' in `List', this is signalled by the value
%%	`Default'. By default, `Default' is the atom `undefined'.

get_value(Key, List) ->
    get_value(Key, List, undefined).

get_value(Key, [P | Ps], Default) ->
    if atom(P), P =:= Key ->
	    true;
       tuple(P), element(1, P) =:= Key ->
	    case P of
		{_, Value} ->
		    Value;
		_ ->
		    %% Don't continue the search!
		    undefined
	    end;
       true ->
	    get_value(Key, Ps, Default)
    end;
get_value(Key, [], Default) ->
    Default.


%% =====================================================================
%% get_bool(Key, List) -> bool()
%%
%%	    Key = Name | term()
%%	    Name = atom()
%%	    List = [Name | {Key, ...} | term()]
%%
%%	If `lookup(Key, List)' would yield `{Key, true}', this function
%%	returns `true'; otherwise `false' is returned. Thus, the entry
%%	associated with `Key' is assumed to be a 2-tuple whose second
%%	element is a boolean value (`true' or `false'), and if it is
%%	not, or if there is no entry associated with `Key' in `List',
%%	the value `false' is used.

get_bool(Key, [P | Ps]) ->
    if atom(P), P =:= Key ->
	    true;
       tuple(P), element(1, P) =:= Key ->
	    case P of
		{_, true} ->
		    true;
		_ ->
		    %% Don't continue the search!
		    false
	    end;
       true ->
	    get_bool(Key, Ps)
    end;
get_bool(Key, []) ->
    false.


%% =====================================================================
%% enter(Property, List) -> List1
%% enter(Key, Value, List) -> List1
%%
%%	    Property = Name | {Key, ...} | term()
%%	    Name = atom()
%%	    Key = Name | term()
%%	    List = List1 = [Property]
%%	    Value = term()
%%
%%	`enter(Property, List)' adds an entry corresponding to
%%	`Property' into the list `List', replacing any existing entry
%%	for the same key.
%%
%%	`enter(Key, Value, List)' has the same effect as `enter({Key,
%%	Value}, List)'.

enter(Property, Ps) ->
    if atom(Property) ->
	    enter_1(Property, Property, Ps);
       tuple(Property), size(Property) >= 1 ->
	    enter_1(element(1, Property), property(Property), Ps);
       true ->
	    [Property | Ps]
    end.

enter(Key, Value, Ps) ->
    enter_1(Key, property(Key, Value), Ps).

%% Entry is an atom or a tuple of arity >= 1. Only if the key is
%% an atom may it occur "naked" in the list.

enter_1(Key, Entry, [P | Ps]) ->
    if atom(P), P =:= Key ->
	    [Entry | Ps];
       tuple(P), Key =:= element(1, P) ->
	    [Entry | Ps];
       true ->
	    [P | enter_1(Key, Entry, Ps)]
    end;
enter_1(_, Entry, []) ->
    [Entry].


%% =====================================================================
%% delete(Key, List) -> List1
%%
%%	    Key = Name | term()
%%	    Name = atom()
%%	    List = List1 = [Name | {Key, ...} | term()]
%%
%%	Deletes all occurrences of `{Key, ...}', and if `Key' is an
%%	atom, also all occurrences of `Key', from `List'

delete(Key, [P | Ps]) ->
    if atom(P), P =:= Key ->
	    delete(Key, Ps);
       tuple(P), element(1, P) =:= Key ->
	    delete(Key, Ps);
       true ->
	    [P | delete(Key, Ps)]
    end;
delete(_, []) ->
    [].


%% =====================================================================
%% concat(List1, List2) -> List
%%
%%	    List1 = List2 = List = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	Joins the property lists `List1' and `List2' by inserting the
%%	elements of `List1' in reverse order into `List2'. This is less
%%	efficient than ordinary list concatenation, but may result in a
%%	shorter list, since all entries that are overridden by some
%%	entry in `List1' are removed.
%%
%% compact(List) -> List1
%%
%%	    List = List1 = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	`compact(List)' is equivalent to `concat(List, [])', which will
%%	remove any and all redundant entries from `List', and use the
%%	most compact representation for each entry.

concat([P | Ps], List) ->
    enter(P, concat(Ps, List));
concat([], List) ->
    List.

compact(Ps) ->
    concat(Ps, []).


%% =====================================================================
%% unfold(List) -> [{Key, ...} | term()]
%%
%%	    List = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	Unfolds all occurences of atoms A in `List' to 2-tuples `{A,
%%	true}'. The function `compact' can be used to turn the list back
%%	into a compact representation.

unfold([P | Ps]) ->
    if atom(P) ->
	    [{P, true} | unfold(Ps)];
       true ->
	    [P | unfold(Ps)]
    end;
unfold([]) ->
    [].


%% =====================================================================
%% map(Function, List) -> List1
%%
%%	    Function = (term()) -> term()
%%	    List = List1 = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	Applies `Function' on each entry in `List' yielding `List1'. For
%%	atoms `A' in `List', `{A, true}' is passed to the function; the
%%	returned values are stored on normal form in `List1'.

map(F, [P | Ps]) ->
    P1 = if atom(P) ->
		 {P, true};
	    true ->
		 P
	 end,
    [property(F(P1)) | map(F, Ps)];
map(_, []) ->
    [].


%% =====================================================================
%% filter(Predicate, List) -> List1
%%
%%	    Predicate = (term()) -> bool()
%%	    List = List1 = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	Removes all entries in `List' for which `Predicate' yields
%%	`false', yielding `List1'. For atoms `A' in `List', `{A, true}'
%%	is passed to the predicate.

filter(F, [P | Ps]) ->
    P1 = if atom(P) ->
		 {P, true};
	    true ->
		 P
	 end,
    case F(P1) of
	true ->
	    [P | filter(F, Ps)];
	false ->
	    filter(F, Ps)
    end;
filter(_, []) ->
    [].


%% =====================================================================
%% get_keys(List) -> [Key]
%%
%%	    Expansions = [{Property, Expansion}]
%%	    List = [Property]
%%	    Property = Name | {Key ...} | term()
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	Returns a list of the keys used in `List', without duplicates.

get_keys(Ps) ->
    sets:to_list(get_keys(Ps, sets:new())).

get_keys([P | Ps], Keys) ->
    if atom(P) ->
	    get_keys(Ps, sets:add_element(P, Keys));
       tuple(P), size(P) >= 1 ->
	    get_keys(Ps, sets:add_element(element(1, P), Keys));
       true ->
	    get_keys(Ps, Keys)
    end;
get_keys([], Keys) ->
    Keys.


%% =====================================================================
%% substitute_aliases(Aliases, List) -> List1
%%
%%	    Aliases = [{Key1, Key2}]
%%	    List = List1 = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key1 = Key2 = Key = Name | term()
%%
%%	Replaces each entry in `List' associated with some key `Key1' in
%%	`Aliases' with a corresponding entry associated with `Key2'. If
%%	the same value of `Key1' occurs more than once in `Aliases',
%%	only the first occurrence is used. For example,
%%	`substitute_aliases([{colour, color}], L)' will replace any
%%	tuple `{colour, ...}' in `L' with `{color, ...}', and any atom
%%	`colour' with `color'.

substitute_aliases(As, [P | Ps]) ->
    [substitute_aliases_1(As, P) | substitute_aliases(As, Ps)];
substitute_aliases(As, []) ->
    [].

substitute_aliases_1([{Key, Key1} | As], P) ->
    if atom(P), P =:= Key ->
	    property(Key1, true);
       tuple(P), element(1, P) =:= Key ->
	    property(setelement(1, P, Key1));
       true ->
	    substitute_aliases_1(As, P)
    end;
substitute_aliases_1([], P) ->
    P.


%% =====================================================================
%% substitute_negations(Aliases, List) -> List1
%%
%%	    Aliases = [{Key1, Key2}]
%%	    List = List1 = [Name | {Key, ...} | term()]
%%	    Name = atom()
%%	    Key1 = Key2 = Key = Name | term()
%%
%%	Replaces each entry in `List' associated with some key `Key1' in
%%	`Aliases', which is assumed to be a boolean property, with a
%%	corresponding *negated* entry associated with `Key2'. If the
%%	same value of `Key1' occurs more than once in `Aliases', only
%%	the first occurrence is used. For example,
%%	`substitute_negations([{no_foo, foo}], L)' will replace any
%%	tuple `{no_foo, X}' in `L' with the value of `{foo, not X}', and
%%	any atom `no_foo' with `{foo, false}'.

substitute_negations(As, [P | Ps]) ->
    [substitute_negations_1(As, P) | substitute_negations(As, Ps)];
substitute_negations(As, []) ->
    [].

substitute_negations_1([{Key, Key1} | As], P) ->
    if atom(P), P =:= Key ->
	    property(Key1, false);
       tuple(P), element(1, P) =:= Key ->
	    case P of
		{_, true} ->
		    property(Key1, false);
		{_, false} ->
		    property(Key1, true);
		_ ->
		    %% The property is supposed to be a boolean, so any
		    %% other tuple is interpreted as `false', as done in
		    %% `get_bool'.
		    property(Key1, true)
	    end;		    
       true ->
	    substitute_negations_1(As, P)
    end;
substitute_negations_1([], P) ->
    P.


%% =====================================================================
%% expand(Expansions, List) -> List1
%%
%%	    Expansions = [{Property, Expansion}]
%%	    List = List1 = Expansion = [Property]
%%	    Property = Name | {Key ...} | term()
%%	    Name = atom()
%%	    Key = Name | term()
%%
%%	For each element `{Property_i, Expansion_i}' in `Expansions', in
%%	order, let `E_i' be the first entry in `List' with the same key
%%	as `Property_i' (if one exists). `List1' is then `List' with the
%%	elements of each `Expansion_i' inserted in place of the
%%	corresponding `E_i', whenever `E_i' and `Property_i' have
%%	equivalent normal forms. If an entry is expanded, then any
%%	following entries with the same key are deleted from the list.
%%
%%	For example, all of `expand([{foo, [bar, baz]}], [fie, foo,
%%	fum])', `expand([{{foo, true}, [bar, baz]}], [fie, foo, fum]'
%%	and `expand([{{foo, false}, [bar, baz]}], [fie, {foo, false},
%%	fum]' return `[fie, bar, baz, fum]'. However, no expansion is
%%	done in the call `expand([{{foo, true}, [bar, baz]}], [fie,
%%	{foo, false}, foo, fum])', because `{foo, false}' shadows `foo'.
%%
%%	Note that if the existing property is to be preserved when
%%	expanded, it must be included in the expansion list. The
%%	inserted expansions are not expanded recursively.

expand(Es, Ps) ->
    flatten(expand_0(key_uniq(normalize(Es)), Ps)).

%% Here, all key properties are normalized and there are no multiple
%% entries in the list of expansions for any specific key property. We
%% insert the expansions one at a time - this is quadratic, but gives
%% the desired behaviour in a simple way.

expand_0([{P, L} | Es], Ps) ->
    expand_0(Es, expand_1(P, L, Ps));
expand_0([], Ps) ->
    Ps.

expand_1(P, L, Ps) ->
    %% First, we must find out what key to look for.
    %% P has a minimal representation here.
    if atom(P) ->
	    expand_2(P, P, L, Ps);
       tuple(P), size(P) >= 1 ->
	    expand_2(element(1, P), P, L, Ps);
       true ->
	    Ps    % refuse to expand non-property
    end.

expand_2(Key, P1, L, [P | Ps]) ->
    if atom(P), P =:= Key ->
	    expand_3(Key, P1, P, L, Ps);
       tuple(P), element(1, P) =:= Key ->
	    expand_3(Key, P1, property(P), L, Ps);
       true ->
	    %% This case handles non-property entries, and thus
	    %% any already inserted expansions (lists), by simply
	    %% ignoring them.
	    [P | expand_2(Key, P1, L, Ps)]
    end;
expand_2(_, _, _, []) ->
    [].

expand_3(Key, P1, P, L, Ps) ->
    %% Here, we have found the first entry with a matching key. Both P
    %% and P1 have minimal representations here. The inserted list will
    %% be flattened afterwards. If the expansion is done, we drop the
    %% found entry and alao delete any later entries with the same key.
    if P1 =:= P ->
	    [L | delete(Key, Ps)];
       true ->
	    %% The existing entry does not match - keep it.
	    [P | Ps]
    end.

normalize([{P, V} | Ps]) ->
    [{property(P), V} | normalize(Ps)];
normalize([]) ->
    [].

key_uniq([{K, V} | Ps]) ->
    [{K, V} | key_uniq_1(K, Ps)];
key_uniq([]) ->
    [].

key_uniq_1(K, [{K1, V} | Ps]) ->
    if K =:= K1 ->
	    key_uniq_1(K, Ps);
       true ->
	    [{K1, V} | key_uniq_1(K1, Ps)]
    end;
key_uniq_1(K, []) ->
    [].

%% This does top-level flattening only.

flatten([E | Es]) when list(E) ->
    E ++ flatten(Es);
flatten([E | Es]) ->
    [E | flatten(Es)];
flatten([]) ->
    [].


%% =====================================================================
%% normalize(Aliases, ExpansionLists, Negations, List) -> List1
%%
%%	    Aliases = Negations = [{Key1, Key2}]
%%	    Key1 = Key2 = Key = Name | term()
%%	    ExpansionLists = [[{Property, Expansion}]]
%%	    Property = Name | {Key ...} | term()
%%	    Name = atom()
%%	    List = List1 = Expansion = [Property]
%%
%%	This passes `List' through all the typical expansion stages.
%%	First, `substitute_aliases' is applied using the `Aliases'
%%	argument; then, for each list in `ExpansionLists', the `expand'
%%	function is applied; finally, `substitute_negations' is applied
%%	using the `Negations' argument, and the result is compacted (cf.
%%	`compact').

normalize(Aliases, ExpansionLists, Negations, L0) ->
    L1 = substitute_aliases(Aliases, L0),
    L2 = do_expansions(ExpansionLists, L1),
    L3 = substitute_negations(Negations, L2),
    compact(L3).

do_expansions([E | Es], L) ->
    L1 = property_lists:expand(E, L),
    do_expansions(Es, L1);
do_expansions([], L) ->
    L.


%% =====================================================================
